var AmlHints = new (function () {
  var _inn;
  var _soapAction;
  var metadataCache = { types : {}, methods : {} };
  
  var getPath = function(xml) {
    var result = {
      path : new Array(),
      completionType : 'element',
      completionName : null
    }
    
    var parser = sax.parser(false, {lowercase: true});
    parser.onerror = function (e) {
      this.completionType = 'element';
      if (this.tagName) {
        var curr = result.path.pop();
        if (!curr || curr._name) {
          if ('undefined' !== typeof (curr)) result.path.push(curr);
          curr = {};
        }
        curr._name = this.tagName;
        result.path.push(curr);

        if (this.attribName) {
          result.completionType = 'attr_value';
          result.completionName = this.attribName;
        } else {
          result.completionType = 'attr';
        }
      }
    };
    parser.onopentag = function (node) {
      var curr = result.path.pop();
      if (!curr || curr._name) {
        if ('undefined' !== typeof (curr)) result.path.push(curr);
        curr = {};
      }
      curr._name = node.name;
      result.path.push(curr);
    };
    parser.onclosetag = function (node) {
      result.path.pop();
    }
    parser.onattribute = function (attr) {
      if (attr.name === 'type' || attr.name === 'action') {
        var curr = result.path.pop();
        if (!curr || curr._name) {
          if ('undefined' !== typeof (curr)) result.path.push(curr);
          curr = {};
        }
        curr[attr.name] = attr.value;
        result.path.push(curr);
      }
    };
    parser.ontext = function (text) {
      this.completionType = 'text';
    };
    
    parser.write(xml).close();
    
    return result;
  }
  
  var ensureProps = function(type, callback) {
    var typeData = metadataCache.types[type];
    var hasProps = false;
    if (typeData && typeData.id) {
      if (typeData.props) {
        for (k in typeData.props) {
          if (Object.prototype.hasOwnProperty.call(typeData.props, k)) {
            hasProps = true;
            break;
          }
        }
      }
      if (!hasProps) {
        _inn.applyAML('<AML><Item action="get" type="Property" select="name,data_type,data_source" related_expand="0"><source_id>' + typeData.id + '</source_id></Item></AML>', function(results) {
          IOM.ItemHelper.forEach(results, function(result) {
            typeData.props[result.getProperty('name')] = { type : result.getProperty('data_type'), source : result.getPropertyAttribute('data_source', 'keyed_name') };
          });
          if ('undefined' !== typeof callback) callback(typeData);
        });
      }
    }
  }

  var selectPath = function (selectStr) {
    var lastOperator=-1;
    var lastIdentifier;
    var char;
    var path = [];
    for(var i=0;i<selectStr.length;i++) {
      char = selectStr.charAt(i);
      if (char === '(' || char === ')' || char === ',') {
        lastIdentifier = selectStr.slice(lastOperator + 1, i);
        lastOperator = i;
        switch (char) {
          case '(':
            path.push(trim(lastIdentifier));
            break;
          case ')':
            path.pop();
            break;
        }
      }
    }
    path.push(trim(selectStr.slice(lastOperator + 1)));
    return path;
  }
  
  var trim = function (value) {
    if ('undefined' === typeof value.trim) {
      return value.replace(/^[\x20\t\r\n\f]+|((?:^|[^\\])(?:\\.)*)[\x20\t\r\n\f]+$/gi, '$1');
    } else {
      return value.trim();
    }
  }
  
  this.soapAction = function (soapAction) {
    _soapAction = soapAction;
  };
  
  this.init = function(inn, soapAction) {
    var getNewTypeMetadata = function () { return { id: null, props : {}, rels : {} }; };
    metadataCache = { types : {}, methods : {} };
    
    _inn = inn;
    _soapAction = soapAction;
    var newData;
    _inn.applyAML('<AML><Item action="get" type="ItemType" select="name" /><Item action="get" type="RelationshipType" related_expand="0" select="related_id,source_id,relationship_id,name" /></AML>', function(results) {
      IOM.ItemHelper.forEach(results, function(result) {
        switch (result.getType())
        {
          case 'RelationshipType':
            var typeData = metadataCache.types[result.getPropertyAttribute('source_id', 'keyed_name')];
            if ('undefined' === typeof (typeData)) typeData = getNewTypeMetadata();
            typeData.rels[result.getProperty('name')] = true;
            break;
          default: //'ItemType'
            if (!metadataCache.types[result.getProperty('name')]) {
              newData = getNewTypeMetadata();
              newData.id = result.getID();
              metadataCache.types[result.getProperty('name')] = newData;
            }
            break;
        }
      });
    });
    _inn.applyAML('<AML><Item action="get" type="Method" select="name" /></AML>', function(results) {
      IOM.ItemHelper.forEach(results, function(result) {
        metadataCache.methods[result.getProperty('name')] = true;
      });
    });
  }
  this.getHints = function(cm) {
    var cursor = cm.getCursor();

    if (cursor.ch > 0) {
      //Separate the text into the portions before and after the last trigger
      var text = cm.getRange(CodeMirror.Pos(0, 0), cursor);
      var typed = '';
      
      var quoteCount = 0;
      for(var i = text.length - 1; i >= 0; i--) {
        if (text[i] == '"') {
          quoteCount++;
        } else if (text[i] == '<') {
          break;
        }
      }
      for(var i = text.length - 1; i >= 0; i--) {
        if((quoteCount % 2 === 0 && (text[i] == ' ' || text[i] == '<')) || 
           (quoteCount % 2 === 1 && text[i] == '"')) {
          typed = text.slice(i+1).toLowerCase();
          text = text.slice(0, i+1);
          break;
        }
      }

      // Get the completion data
      var complData = getPath(text);
            
      // Ensure properties are loaded for the current type and subtypes.
      if (complData.path.length > 0 && complData.path[complData.path.length - 1]._name === 'item' && complData.path[complData.path.length - 1].type) {
        ensureProps(complData.path[complData.path.length - 1].type, function (type) {
          var loadTypes = {};
          for (prop in type.props) {
            if (type.props[prop].type === "item" && type.props[prop].source !== complData.path[complData.path.length - 1].type
                && 'undefined' !== typeof type.props[prop].source) {
              loadTypes[type.props[prop].source] = true;
            }
          }
          for (loadType in loadTypes) {
            ensureProps(loadType);
          }
        });
      }
      
      // Calculate the hints
      var hints = [];
      var last = complData.path[complData.path.length - 1];
      switch (complData.completionType) {
        case 'attr_value':
          if (last._name === 'item') {
            switch (complData.completionName) {
              case 'action':
                hints.push('ActivateActivity"');
                hints.push('add"');
                hints.push('AddItem"');
                hints.push('AddHistory"');
                hints.push('ApplyUpdate"');
                hints.push('BuildProcessReport"');
                hints.push('CancelWorkflow"');
                hints.push('checkImportedItemType"');
                hints.push('closeWorkflow"');
                hints.push('copy"');
                hints.push('copyAsIs"');
                hints.push('copyAsNew"');
                hints.push('create"');
                hints.push('delete"');
                hints.push('edit"');
                hints.push('EmailItem"');
                hints.push('EvaluateActivity"');
                hints.push('exportItemType"');
                hints.push('get"');
                hints.push('getItemAllVersions"');
                hints.push('getAffectedItems"');
                hints.push('getItemConfig"');
                hints.push('getItemLastVersion"');
                hints.push('getItemNextStates"');
                hints.push('getItemRelationships"');
                hints.push('GetItemRepeatConfig"');
                hints.push('getItemWhereUsed"');
                hints.push('GetMappedPath"');
                hints.push('getPermissions"');
                hints.push('getRelatedItem"');
                hints.push('GetUpdateInfo"');
                hints.push('instantiateWorkflow"');
                hints.push('lock"');
                hints.push('merge"');
                hints.push('New Workflow Map"');
                hints.push('PromoteItem"');
                hints.push('purge"');
                hints.push('recache"');
                hints.push('replicate"');
                hints.push('resetAllItemsAccess"');
                hints.push('resetItemAccess"');
                hints.push('resetLifecycle"');
                hints.push('setDefaultLifecycle"');
                hints.push('skip"');
                hints.push('startWorkflow"');
                hints.push('unlock"');
                hints.push('update"');
                hints.push('ValidateWorkflowMap"');
                hints.push('version"');
                
                if (metadataCache.methods) {
                  for (method in metadataCache.methods) {
                    hints.push(method + '"');
                  }
                }
                break;
              case 'access_type':
                hints.push('can_add"');
                hints.push('can_delete"');
                hints.push('can_get"');
                hints.push('can_update"');
                break;
              case 'doGetItem':
              case 'version':
              case 'isCriteria':
              case 'related_expand':
              case 'serverEvents':
                hints.push('0"');
                hints.push('1"');
                break;
              case 'queryType':
                hints.push('Effective"');
                hints.push('Latest"');
                hints.push('Released"');
                break;
              case 'orderBy':
                if (last.type) {
                  var lastComma = typed.lastIndexOf(',') + 1;
                  if (lastComma > 0) {
                    text += typed.slice(0, lastComma);
                    typed = typed.slice(lastComma);
                  }
                  
                  var typeData = metadataCache.types[last.type];
                  if (typeData && typeData.props) {
                    for (prop in typeData.props) {
                      hints.push(prop);
                      hints.push(prop + ' DESC');
                    }
                  }
                }
                break;
              case 'select':
                if (last.type) {
                  var path = selectPath(typed);
                  text += typed.slice(0, typed.length - path[path.length - 1].length);
                  typed = typed.slice(typed.length - path[path.length - 1].length);
                  
                  var currType = last.type;
                  var currProp;
                  for(var i = 0; i < path.length - 1; i++) {
                    currProp = metadataCache.types[currType].props[path[i]];
                    if (currProp && currProp.type === "item") {
                      currType = currProp.source;
                    } else {
                      break;
                    }
                  }
                  
                  for (prop in metadataCache.types[currType].props) {
                    hints.push(prop);
                  }
                }
                break;
              case 'type':
                if (complData.path.length > 2 && complData.path[complData.path.length - 3]._name === 'item' &&
                   complData.path[complData.path.length - 2]._name === 'relationships') {
                  for (rel in metadataCache.types[complData.path[complData.path.length - 3].type].rels) {
                    hints.push(rel + '"');
                  }
                } else {
                  for (name in metadataCache.types) {
                    hints.push(name + '"');
                  }
                }
                break;
              case 'where':
                if (last.type) {
                  var typeData = metadataCache.types[last.type];
                  if (typeData && typeData.props) {
                    for (prop in typeData.props) {
                      hints.push('[' + last.type.replace(/\s/gi, '_') + '].' + prop);
                    }
                  }
                }
                break;
            }
          } else {
            switch (complData.completionName) {
              case 'condition':
                hints.push('between"');
                hints.push('eq"');
                hints.push('ge"');
                hints.push('gt"');
                hints.push('in"');
                hints.push('is not null"');
                hints.push('is null"');
                hints.push('is"');
                hints.push('le"');
                hints.push('like"');
                hints.push('lt"');
                hints.push('ne"');
                hints.push('not between"');
                hints.push('not in"');
                hints.push('not like"');
                break;
              case 'is_null':
                hints.push('0"');
                hints.push('1"');
                break;
            }
          }
          break;
        case 'attr':
          switch (last._name) {
            case 'item':
              switch (_soapAction) {
                case 'GenerateNewGUIDEx':
                  hints.push('quantity=');
                  break;
                case 'GenerateNewGUID':
                  break;
                default:
                  hints.push('action=');
                  if (last.action === 'getPermissions') hints.push('access_type=');
                  hints.push('config_path=');
                  hints.push('doGetItem=');
                  hints.push('id=');
                  hints.push('idlist=');
                  hints.push('isCriteria=');
                  hints.push('language=');
                  hints.push('levels=');
                  hints.push('maxRecords=');
                  hints.push('page=');
                  hints.push('pagesize=');
                  hints.push('orderBy=');
                  hints.push('queryDate=');
                  hints.push('queryType=');
                  hints.push('related_expand=');
                  hints.push('select=');
                  hints.push('serverEvents=');
                  hints.push('type=');
                  hints.push('typeID=');
                  hints.push('version=');
                  hints.push('where=');
                  break;
              }
              break;
            case 'aml':
            case 'and':
            case 'not':
            case 'or':
            case 'relationships':
              break;
            default:
              hints.push('condition=');
              hints.push('is_null=');
              break;
          }
          break;
        default: //'element'
          if (complData.path.length === 0) {
            switch (_soapAction) {
              case 'ApplySQL':
                hints.push('sql');
                break;
              case 'ApplyAML':
                hints.push('AML');
                // break purposely omitted
              default:
                hints.push('Item');
                break;
            }
          } else if (complData.path.length === 1 && complData.path[0]._name === 'aml') {
            hints.push('Item');
          } else {
            var j = complData.path.length - 1;
            while (complData.path[j]._name === 'and' || complData.path[j]._name === 'not' || complData.path[j]._name === 'or') j--;
            var last = complData.path[j];
            if (last._name === 'item') {
              hints.push('Relationships');
              if (last.action === 'get') {
                hints.push('and');
                hints.push('not');
                hints.push('or');
              }
              if (last.type) {
                var typeData = metadataCache.types[last.type];
                if (typeData && typeData.props) {
                  for (prop in typeData.props) {
                    hints.push(prop);
                  }
                }
              }
            } else if (complData.path.length > 1) {
              var lastItem = null;
              for (var i = complData.path.length - 2; i >= 0; i--) {
                if (complData.path[i]._name === 'item') {
                  lastItem = complData.path[i];
                  break;
                }
              }
              
              if (lastItem) {
                if (last._name === "relationships") {
                  hints.push('Item');
                } else {
                  var prop = metadataCache.types[lastItem.type].props[last._name];
                  if (prop && prop.type === "item" && prop.source !== lastItem.type) {
                    hints.push('Item type="' + prop.source  + '"');
                  }
                }
              }
            }
            
            if (hints.length === 0 && last._name === 'result') hints.push('Item');
              
            hints.sort(
              function(a, b) {
                if (a.toLowerCase() < b.toLowerCase()) return -1;
                if (a.toLowerCase() > b.toLowerCase()) return 1;
                return 0;
              }
            );
          }
          break;
      }

      var findIndex;
      var partials = [];
      for (var i = hints.length - 1; i >= 0; i--) {
        findIndex = hints[i].toLowerCase().indexOf(typed);
        if(findIndex > 0) partials.push(hints[i]);
        if(findIndex !== 0) hints.splice(i, 1);
      }
      for (var i = partials.length - 1; i >= 0; i--) {
        hints.push(partials[i]);
      }
      
      return {
          list: hints,
          from: CodeMirror.Pos(cursor.line, cursor.ch - typed.length),
          to: cursor
      };
    }
  }
})();